/* ------------------------------------------------------------------------ */
/* matrix.c - Funktionsbibliothek zur Arbeit mit Vektoren und Matrizen      */
/*                                                                          */
/*               Copyright Tilman Kuepper 2009 - 2020.                      */
/*    Distributed under the Boost Software License, Version 1.0.            */
/*        (See accompanying file LICENSE_1_0.txt or copy at                 */
/*               http://www.boost.org/LICENSE_1_0.txt)                      */
/*                                                                          */
/* Letzte Bearbeitung: 5. Juli 2020                                         */
/* ------------------------------------------------------------------------ */

#undef  _CRT_SECURE_NO_WARNINGS 
#define _CRT_SECURE_NO_WARNINGS  /* Warnungen bei MSVC++ ausschalten */

/* ------------------------------------------------------------------------ */
/* Beim Schreiben/Lesen von CSV-Dateien wird versucht, das landestypische   */
/* Zahlenformat (Dezimalpunkt, Dezimalkomma) zu verwenden. Dazu wird die    */
/* Funktion "setlocale" aus der Header-Datei "locale.h" verwendet. Falls    */
/* dies nicht erwuenscht ist, kann "MAT_USE_LOCALE" auf 0 gesetzt werden.   */
/* ------------------------------------------------------------------------ */
#ifndef MAT_USE_LOCALE
#define MAT_USE_LOCALE 1
#endif

#define _USE_MATH_DEFINES  /* Microsoft: M_PI usw... */

#include "matrix.h"
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <stdarg.h>

#if MAT_USE_LOCALE
#include <locale.h>
#endif

/* ------------------------------------------------------------------------ */
/* Rueckgabewerte bzw. Fehlermeldungen in Textform.                         */
/* ------------------------------------------------------------------------ */
const char *mat_errmsg[] =
{
    "OK",                           /* MAT_OK         */
    "Matrix inversion failed",      /* MAT_ERR_NOINV  */
    "Cannot open CSV file",         /* MAT_ERR_FOPEN  */
    "Cannot write CSV file",        /* MAT_ERR_FWRITE */
    "Invalid matrix format",        /* MAT_ERR_FORMAT */
    "Out of memory",                /* MAT_ERR_MEMORY */
    "Matrices are not equal",       /* MAT_ERR_NOTEQ  */
    "Invalid dimension",            /* MAT_ERR_DIM    */
    "Vector normalization failed",  /* MAT_ERR_VNORM  */
    "Iteration failed"              /* MAT_ERR_ITFAIL */
};

/* ------------------------------------------------------------------------ */
/* Definitionen zur Ausgabe von Matrizen und Vektoren                       */
/* ------------------------------------------------------------------------ */
#ifndef MAT_CSV_FMT
#define MAT_CSV_FMT "%.20g"  /* Ausgabe in CSV-Dateien */
#endif

#ifndef MAT_PRN_FMT
#define MAT_PRN_FMT "%10.2f"  /* Ausgabe auf dem Bildschirm */
#endif

/* ------------------------------------------------------------------------ */
/* Makros zum einfachen Zugriff auf einzelne Matrix-Elemente.               */
/* ------------------------------------------------------------------------ */
#undef MAT
#undef MAT1
#undef MAT2
#undef R
#undef Q
#undef T
#undef Z
#undef LU
#undef RESULT
#undef TMP
#undef TMP1
#undef TMP2

#define MAT(row, col)     ((double*)mat   )[(col) + dim * (row)]
#define MAT1(row, col)    ((double*)mat1  )[(col) + dim * (row)]
#define MAT2(row, col)    ((double*)mat2  )[(col) + dim * (row)]
#define R(row, col)       ((double*)r     )[(col) + dim * (row)]
#define Q(row, col)       ((double*)q     )[(col) + dim * (row)]
#define T(row, col)       ((double*)t     )[(col) + dim * (row)]
#define Z(row, col)       ((double*)z     )[(col) + dim * (row)]
#define LU(row, col)      ((double*)lu    )[(col) + dim * (row)]
#define RESULT(row, col)  ((double*)result)[(col) + dim * (row)]
#define TMP(row, col)     ((double*)tmp   )[(col) + dim * (row)]
#define TMP1(row, col)    ((double*)tmp1  )[(col) + dim * (row)]
#define TMP2(row, col)    ((double*)tmp2  )[(col) + dim * (row)]

/* ------------------------------------------------------------------------ */
/* Makros fuer verschiedene einfache Berechnungen.                          */
/* ------------------------------------------------------------------------ */
#undef NEG_POW
#undef MAT_MAX

#define NEG_POW(p) (((int)(p) % 2) ? -1.0 : +1.0)      /* (-1)^p berechnen  */
#define MAT_MAX(a, b) ((a) > (b) ? (a) : (b))          /* Maximum bestimmen */

/* ------------------------------------------------------------------------ */
/* Mit den folgenden Makros koennen Anweisungen/Befehle/Funktionsaufrufe    */
/* zu Bloecken zusammengefasst werden, die - aehnlich wie bei Schleifen -   */
/* mit "break" verlassen werden koennen. Die Programmausfuehrung wird in    */
/* diesem Fall mit dem ersten Befehl nach dem Ende des Blocks fortgesetzt.  */
/* ------------------------------------------------------------------------ */
#undef BEGIN_BLOCK
#undef END_BLOCK

#define BEGIN_BLOCK for(;;) {
#define END_BLOCK break; }

/* ------------------------------------------------------------------------ */
/* Interne Hilfsfunktionen: Dynamischen Speicher anfordern.                 */
/* ------------------------------------------------------------------------ */

/* Variante fuer Matrizen */
static inline double* mat_alloc(size_t dim)
{
    double* result = (double*)malloc(sizeof(double) * dim * dim);
    assert(result != NULL);
    return result;
}

/* Variante fuer Vektoren */
static inline double* vec_alloc(size_t dim)
{
    double* result = (double*)malloc(sizeof(double) * dim);
    assert(result != NULL);
    return result;
}

/* ------------------------------------------------------------------------ */
/* Interne Hilfsfunktionen: Dynamisch angeforderten Speicher freigeben.     */
/* ------------------------------------------------------------------------ */

/* Variante fuer Matrizen */
static inline void mat_free(double *pmatrix, ...)
{
    va_list list;
    double *p;

    va_start(list, pmatrix); 
    assert(pmatrix != NULL);
    for (p = pmatrix; p != NULL; p = va_arg(list, double*)) free(p);
    va_end(list);
}

/* Variante fuer Vektoren */
static inline void vec_free(double *pvec, ...)
{
    va_list list;
    double *p;

    va_start(list, pvec); 
    assert(pvec != NULL);
    for (p = pvec; p != NULL; p = va_arg(list, double*)) free(p);
    va_end(list);
}

/* ------------------------------------------------------------------------ */
/* Binomialkoeffizient "n ueber k" (siehe m_init, inverse Hilbert-Matrix).  */
/* ------------------------------------------------------------------------ */
double bin_coef_(size_t n, size_t k)
{
  size_t i;
  double ret = 1;
  if (2 * k < n) k = n - k;
  for(i = 1; i <= n - k; i++) ret = ret * (i + k) / i;
  return ret;
}

/* ------------------------------------------------------------------------ */
/* Makros und Hilfsfunktionen zum Rechnen mit komplexen Zahlen.             */
/* ------------------------------------------------------------------------ */

/* Zugriff auf Real- und Imaginaerteil: In der "oeffentlichen" Headerdatei
 * matrix.h gibt es zu diesem Zweck die beiden Makros MAT_REAL und MAT_IMAG.
 * Innerhalb von matrix.c werden zur Verbesserung der Uebersichtlichkeit
 * allerdings die kuerzeren Makros RE und IM verwendet. */
#undef RE
#undef IM

#define RE(z) z.val[0]
#define IM(z) z.val[1]

/* Zwei komplexe Zahlen addieren */
static HM_COMPLEX cadd(HM_COMPLEX a, HM_COMPLEX b)
{
    HM_COMPLEX result;
    RE(result) = RE(a) + RE(b);
    IM(result) = IM(a) + IM(b);
    return result;
}

/* Zwei komplexe Zahlen subtrahieren */
static HM_COMPLEX csub(HM_COMPLEX a, HM_COMPLEX b)
{
    HM_COMPLEX result;
    RE(result) = RE(a) - RE(b);
    IM(result) = IM(a) - IM(b);
    return result;
}

/* Zwei komplexe Zahlen multiplizieren */
static HM_COMPLEX cmul(HM_COMPLEX a, HM_COMPLEX b)
{
    HM_COMPLEX result;
    RE(result) = RE(a) * RE(b) - IM(a) * IM(b);
    IM(result) = RE(a) * IM(b) + IM(a) * RE(b);
    return result;
}

/* Zwei komplexe Zahlen dividieren */
static HM_COMPLEX cdiv(HM_COMPLEX a, HM_COMPLEX b)
{
    HM_COMPLEX result;
    double denom = RE(b) * RE(b) + IM(b) * IM(b);
    RE(result) = RE(a) * RE(b) + IM(a) * IM(b);
    IM(result) = IM(a) * RE(b) - RE(a) * IM(b);

    assert(denom != 0);
    RE(result) /= denom, IM(result) /= denom;
    return result;
}

/* Betrag einer komplexen Zahl ermitteln */
static double c_abs(HM_COMPLEX a)
{
    return sqrt(RE(a) * RE(a) + IM(a) * IM(a));
}

/* Zwei komplexe Zahlen vertauschen */
static void c_swap(HM_COMPLEX *p1, HM_COMPLEX *p2)
{
    HM_COMPLEX tmp = *p1;
    *p1 = *p2, *p2 = tmp;
}

/* ------------------------------------------------------------------------ */
/* Es wird eine Kopie der uebergebenen Zeichenkette erzeugt. Die Kopie      */
/* muss vom Anwender wieder mittels "free" freigegeben werden.              */
/* ------------------------------------------------------------------------ */
static char *str_duplicate_(const char *str)
{
    char *ptr = NULL;
    if(!str) return NULL;  /* leerer String */

    ptr = (char*)malloc(strlen(str) + 1);
    if(!ptr) { assert(!"str_duplicate: out of memory"); return NULL; }

    strcpy(ptr, str);
    return ptr;
}

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird initialisiert (je nach Parameter "type").          */
/* ------------------------------------------------------------------------ */
void m_init(size_t dim, void *mat, int type)
{
    size_t row, col, l;
    assert(dim > 0);
    assert(mat != NULL);

    switch(type)
    {
    case MAT_INIT_NULL:
        /* Alle Elemente auf null setzen */
        memset(mat, 0, sizeof(double) * dim * dim);
        break;

    case MAT_INIT_IDENTITY:
        /* Matrix auf Einheitsmatrix setzen */
        for(row = 0; row < dim; ++row)
            for(col = 0; col < dim; ++col)
                MAT(row, col) = (row == col);
        break;

    case MAT_INIT_HILBERT:
        /* Matrix auf Hilbertmatrix setzen */
        for(row = 0; row < dim; ++row)
            for(col = 0; col < dim; ++col)
                MAT(row, col) = 1.0 / (1 + row + col);
        break;

    case MAT_INIT_INV_HILBERT:
        /* Matrix auf inv. Hilbertmatrix setzen */
        for(row = 0; row < dim; ++row)
            for(col = 0; col < dim; ++col)
                for(l = MAT_MAX(row, col), MAT(row, col) = 0; l < dim; ++l)
                {
                    /* Zur Implementierung siehe C. Wieners, "An explicit
                     * formula for the inverse of the Hilbert matrix"
                     * http://citeseerx.ist.psu.edu/viewdoc/download?doi=
                     * 10.1.1.24.9271&rep=rep1&type=pdf (04.02.2014) */
                    double sum = NEG_POW(row + col) * (2 * l + 1);
                    sum *= bin_coef_(l + col, col), sum *= bin_coef_(l, col);
                    sum *= bin_coef_(l + row, row), sum *= bin_coef_(l, row);
                    MAT(row, col) += sum;
                }
        break;

    case MAT_INIT_RANDOM:
        /* Zufallszahlen zwischen 0...1 erzeugen */
        for(row = 0; row < dim; ++row)
            for(col = 0; col < dim; ++col)
                MAT(row, col) = (double)rand() / RAND_MAX;
        break;

    case MAT_INIT_ONES:
        /* Alle Elemente auf eins setzen */
        for(row = 0; row < dim; ++row)
            for(col = 0; col < dim; ++col)
                MAT(row, col) = 1.0;
        break;

    default:
        assert(!"invalid type");
    }
}

/* ------------------------------------------------------------------------ */
/* Alle Elemente der Matrix "mat" werden auf den angegebenen Wert gesetzt.  */
/* ------------------------------------------------------------------------ */
void m_fill(size_t dim, void *mat, double value)
{
    size_t row, col;
    assert(dim > 0);
    assert(mat != NULL);

    for(row = 0; row < dim; ++row)
        for(col = 0; col < dim; ++col)
            MAT(row, col) = value;
}

/* ------------------------------------------------------------------------ */
/* Die Elemente der Matrix "mat" werden auf dem Bildschirm ausgegeben.      */
/* Das Format kann ueber die Konstante MAT_PRN_FMT eingestellt werden.      */
/* ------------------------------------------------------------------------ */
void m_print(size_t dim, const void *mat)
{
    size_t row, col;
    assert(dim > 0);
    assert(mat != NULL);

    for(row = 0; row < dim; ++row)
    {
        for(col = 0; col < dim; ++col)
        {
            if(col) printf(" ");
            printf(MAT_PRN_FMT, MAT(row, col));
        }
        printf("\n");
    }            
    printf("\n");
}

/* ------------------------------------------------------------------------ */
/* Der Inhalt der Matrix "mat" wird nach "mresult" kopiert.                 */
/* ------------------------------------------------------------------------ */
void m_copy(size_t dim, void *result, const void *mat)
{
    assert(dim > 0);
    assert(result != NULL);
    assert(mat != NULL);

    memcpy(result, mat, sizeof(double) * dim * dim);
}

/* ------------------------------------------------------------------------ */
/* Die Matrizen "mat1" und "mat2" werden verglichen. Falls beide Matrizen   */
/* identisch sind, wird MAT_OK zurueckgegeben, andernfalls MAT_ERR_NOTEQ.   */
/* ------------------------------------------------------------------------ */
int m_equal(size_t dim, const void *mat1, const void *mat2)
{
    assert(dim > 0);
    assert(mat1 != NULL);
    assert(mat2 != NULL);

    return memcmp(mat1, mat2, dim * dim * sizeof(double))
        ? MAT_ERR_NOTEQ : MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Die Matrizen "mat1" und "mat2" werden multipliziert.                     */
/* Das Ergebnis wird in der Matrix "mresult" abgelegt.                      */
/* ------------------------------------------------------------------------ */
void m_mul(size_t dim, void *result, const void *mat1, const void *mat2)
{
    /* Ab OpenMP 3.0 sind vorzeichenlose Schleifenvariablen erlaubt,
     * in Visual Studio 2017 ist aber nur OpenMP 2.0 integriert. */
    long lrow, ldim = (long)dim;  /* ...daher: long statt size_t */
    const long matrix_size_threshold = 50;

    assert(result != NULL);
    assert(mat1 != NULL);
    assert(mat2 != NULL);
    assert(ldim > 0);

    /* Faktoren und Produkt muessen unterschiedlich sein! */
    assert(mat1 != result);
    assert(mat2 != result);

    /* Kleinere Matrizen werden "ganz klassisch" mit drei
     * verschachtelten for-Schleifen multipliziert. */
    if(ldim < matrix_size_threshold)
    {
        size_t row, col, idx;
        for(row = 0; row != dim; ++row)
        {
            /* Juli 2017: idx- und col-Schleife vertauscht, so ist
             * die Multiplikation "Cache-freundlicher", siehe auch:
             * http://penguin.ewu.edu/~trolfe/MatMult/MatOpt.html */
            for(col = 0; col != dim; ++col)
                RESULT(row, col) = MAT1(row, 0) * MAT2(0, col);
            for(idx = 1; idx != dim; ++idx)
                for(col = 0; col != dim; ++col)
                    RESULT(row, col) += MAT1(row, idx) * MAT2(idx, col);
        }
    }
    else
    {
        /* Juli 2017: Durch Parallelisierung der ersten for-Schleife
         * laesst sich die Geschwindigkeit noch weiter steigern. */
#ifdef _OPENMP
#pragma omp parallel for
#endif
        for(lrow = 0; lrow < ldim; ++lrow)
        {
            size_t col, idx;
            for(col = 0; col != dim; ++col)
                RESULT(lrow, col) = MAT1(lrow, 0) * MAT2(0, col);
            for(idx = 1; idx != dim; ++idx)
                for(col = 0; col != dim; ++col)
                    RESULT(lrow, col) += MAT1(lrow, idx) * MAT2(idx, col);
        }
    }
}

/* ------------------------------------------------------------------------ */
/* Alle Elemente der Matrix "mat" werden mit dem Faktor "f" multipliziert.  */
/* ------------------------------------------------------------------------ */
void m_fmul(size_t dim, void *mat, double f)
{
    size_t row, col;
    assert(dim > 0);
    assert(mat != NULL);

    for(row = 0; row < dim; ++row)
        for(col = 0; col < dim; ++col)
            MAT(row, col) *= f;
}

/* ------------------------------------------------------------------------ */
/* Die Matrizen "mat1" und "mat2" werden addiert. Das Ergebnis              */
/* wird in der Matrix "mresult" abgelegt.                                   */
/* ------------------------------------------------------------------------ */
void m_add(size_t dim, void *result, const void *mat1, const void *mat2)
{
    size_t row, col;
    assert(dim > 0);
    assert(result != NULL);
    assert(mat1 != NULL);
    assert(mat2 != NULL);

    for(row = 0; row < dim; ++row)
        for(col = 0; col < dim; ++col)
            RESULT(row, col) = MAT1(row, col) + MAT2(row, col);
}

/* ------------------------------------------------------------------------ */
/* Das "f-fache" der Matrix "mat" wird zur Matrix "mresult" addiert.        */
/* ------------------------------------------------------------------------ */
void m_fadd(size_t dim, void *result, const void *mat, double f)
{
    size_t row, col;
    assert(dim > 0);
    assert(result != NULL);
    assert(mat != NULL);

    for(row = 0; row < dim; ++row)
        for(col = 0; col < dim; ++col)
            RESULT(row, col) += f * MAT(row, col);
}

/* ------------------------------------------------------------------------ */
/* Die Zeilen "row1" und "row2" der Matrix "mat" werden vertauscht.         */
/* ------------------------------------------------------------------------ */
void m_swap_row(size_t dim, void *mat, size_t row1, size_t row2)
{
    size_t col;
    double tmp;
    assert(dim > 0);
    assert(mat != NULL);
    assert(row1 < dim);
    assert(row2 < dim);

    if(row1 == row2) return;  // Zeile mit sich selbst tauschen...
    for(col = 0; col < dim; ++col)
        tmp = MAT(row1, col),
        MAT(row1, col) = MAT(row2, col),
        MAT(row2, col) = tmp;
}

/* ------------------------------------------------------------------------ */
/* Alle Elemente der Zeile "row" werden mit einem Faktor multipliziert.     */
/* ------------------------------------------------------------------------ */
void m_mul_row(size_t dim, void *mat, size_t row, double factor)
{
    size_t col;
    assert(dim > 0);
    assert(mat != NULL);
    assert(row < dim);

    for(col = 0; col < dim; ++col) MAT(row, col) *= factor;
}

/* ------------------------------------------------------------------------ */
/* Alle Elemente der Spalte "col" werden mit einem Faktor multipliziert.    */
/* ------------------------------------------------------------------------ */
void m_mul_col(size_t dim, void *mat, size_t col, double factor)
{
    size_t row;
    assert(dim > 0);
    assert(mat != NULL);
    assert(col < dim);

    for(row = 0; row < dim; ++row) MAT(row, col) *= factor;
}

/* ------------------------------------------------------------------------ */
/* Von der Zeile "result" wird das "f-fache" der Zeile "sub" abgezogen.     */ 
/* ------------------------------------------------------------------------ */
void m_sub_row(size_t dim, void *mat, size_t result, size_t sub, double f)
{
    size_t col;
    assert(dim > 0);
    assert(mat != NULL);
    assert(sub < dim);
    assert(result < dim);

    for(col = 0; col < dim; ++col)
        MAT(result, col) -= f * MAT(sub, col);
}

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird invertiert. Falls dabei ein Fehler auftritt, ist   */
/* der Rueckgabewert MAT_ERR_MEMORY (nicht genuegend freier Speicher) oder  */
/* MAT_ERR_NOINV (inverse Matrix existiert nicht). Andernfalls wird die     */
/* inverse Matrix in "mat" abgelegt und der Rueckgabewert ist MAT_OK.       */
/* ------------------------------------------------------------------------ */
int m_invert(size_t dim, void *mat)
{
    size_t row, col;
    double *r = mat_alloc(dim), *i = mat_alloc(dim);

    assert(dim > 0);
    assert(mat != NULL);

    /* Es wird zunaechst die QR-Zerlegung von "mat" ermittelt.
     * Es gilt: Q * R * Inv(A) = Ident --> R * Inv(A) = Transp(Q)
     * Die inverse Matrix folgt durch Rueckwaerts-Einsetzen. */
    if(r == NULL || i == NULL || m_qr_decomp(dim, mat, i, r) != MAT_OK)
    { mat_free(r, i, NULL); return MAT_ERR_MEMORY; }

    m_transpose(dim, i);
    for(row = dim - 1; row != (size_t)-1; --row)
    {
        /* Matrix singulaer? */
        if(R(row, row) == 0) { mat_free(r, i, NULL); return MAT_ERR_NOINV; }
        for(col = row + 1; col < dim; ++col)
        {
            double factor = R(row, col) / R(col, col);
            m_sub_row(dim, r, row, col, factor);
            m_sub_row(dim, i, row, col, factor);
        }
    }

    /* Zum Schluss werden die Diagonalelemente normiert */
    for(row = 0; row < dim; ++row)
        m_mul_row(dim, i, row, 1.0 / R(row, row));

    m_copy(dim, mat, i);
    mat_free(r, i, NULL);
    return MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Die Determinante der Matrix "mat" wird berechnet und zurueckgegeben.     */
/* Falls nicht genuegend freier Speicher zur Verfuegung steht, wird als     */
/* Ergebnis 0 zurueckgegeben und *pResult auf MAT_ERR_MEMORY gesetzt,       */
/* andernfalls auf MAT_OK. Es ist moeglich, fuer pResult NULL zu            */
/* uebergeben, falls keine Erfolgs-/Fehlermeldung erforderlich ist.         */
/* ------------------------------------------------------------------------ */
double m_det(size_t dim, const void *mat, int *pResult)
{
    int ret_code;
    double det, *pol = vec_alloc(dim + 1);
    if(!pol) { assert(!"m_det: out of memory"); return MAT_ERR_MEMORY; }

    /* Zunaechst wird das charakteristische Polynom der Matrix "mat"
     * ermittelt. Der letzte Koeffizient dieses Polynoms entspricht
     * bis auf das Vorzeichen der gesuchten Determinanten von "mat" */
    ret_code = m_charpol(dim, mat, pol);
    det = (ret_code == MAT_OK) ? pol[dim] * NEG_POW(dim) : 0;
    if(pResult) *pResult = ret_code;
    vec_free(pol, NULL);
    return det;
}

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird an der Hauptdiagonalen gespiegelt (transponiert).  */
/* ------------------------------------------------------------------------ */
void m_transpose(size_t dim, void *mat)
{
    double tmp;
    size_t row, col;
    assert(dim > 0);
    assert(mat != NULL);

    for(row = 0; row < dim; ++row)
        for(col = 0; col < row; ++col)
            tmp = MAT(row, col),
            MAT(row, col) = MAT(col, row),
            MAT(col, row) = tmp;
}

/* ------------------------------------------------------------------------ */
/* Ist die Matrix "mat" symmetrisch (Rueckgabe = MAT_OK) oder ist sie nicht */
/* symmetrisch (Rueckgabe = MAT_ERR_NOTEQ)? Bei einer symmetrischen Matrix  */
/* weichen die an der Hauptdiagonalen gespiegelten Matrixelemente maximal   */
/* um den Betrag "epsilon" voneinander ab.                                  */
/* ------------------------------------------------------------------------ */
int m_is_symmetric(size_t dim, const void *mat, double epsilon)
{
    size_t row, col;
    assert(dim > 0);
    assert(mat != NULL);

    epsilon = fabs(epsilon);
    for(row = 1; row < dim; ++row)
        for(col = 0; col < row; ++col)
            if(fabs(MAT(row, col) - MAT(col, row)) > epsilon)
                return MAT_ERR_NOTEQ;
    return MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird symmetrisch gemacht. Entweder dadurch, dass        */
/* die Elemente oberhalb der Hauptdiagonalen nach unten kopiert werden      */
/* (mode = MAT_USE_UPPER) oder dadurch, dass die Elemente unterhalb der     */
/* Hauptdiagonalen nach oben kopiert werden (mode = MAT_USE_LOWER) oder     */
/* dadurch, dass die Mittelwerte aller an der Hauptdiagonalen gespiegelten  */
/* Elemente berechnet werden (mode = MAT_USE_MEAN).                         */
/* ------------------------------------------------------------------------ */
void m_make_symmetric(size_t dim, void *mat, int mode)
{
    size_t row, col;
    assert(dim > 0);
    assert(mat != NULL);
    assert(mode >= 0 && mode <= 2);

    double factor_upper, factor_lower;
    switch(mode)
    {
    case MAT_USE_UPPER: factor_upper = 1.0, factor_lower = 0.0; break;
    case MAT_USE_LOWER: factor_lower = 1.0, factor_upper = 0.0; break;
    default:            factor_upper = 0.5, factor_lower = 0.5; break;
    }

    for(row = 1; row < dim; ++row)
        for(col = 0; col < row; ++col)
            MAT(row, col) = MAT(col, row) =
            MAT(row, col) * factor_lower + MAT(col, row) * factor_upper;
}

/* ------------------------------------------------------------------------ */
/* Es wird die QR-Zerlegung der Matrix "mat" ermittelt. Der Rueckgabewert   */
/* ist entweder MAT_ERR_MEMORY (zu wenig freier Speicher) oder MAT_OK.      */
/* ------------------------------------------------------------------------ */
int m_qr_decomp(size_t dim, const void *mat, void *q, void *r)
{
    size_t idx, col, row, k;
    double *tmp = mat_alloc(dim);
    double t = 1.0, s = 1.0, c = 1.0;

    assert(q != NULL); assert(dim > 0);
    assert(r != NULL); assert(mat != NULL);
    if(!tmp) { assert(!"m_qr_decomp: out of memory"); return MAT_ERR_MEMORY; }

    /* "mat", "q" und "r" muessen unterschiedliche Matrizen sein! */
    assert(mat != q); assert(mat != r); assert(q != r);

    /* Matrizen Q und R initialisieren. */
    m_init(dim, q, MAT_INIT_IDENTITY);
    m_copy(dim, r, mat);

    /* Schritt fuer Schritt werden alle Elemente unterhalb der
     * Hauptdiagonalen von R durch Givens-Rotationen beseitigt. */
    for(col = 0; col < dim; ++col)
    {
        for(row = dim - 1; row > col; --row)
        {
            /* Ist das aktuelle Element evtl. schon gleich null? */
            if(R(row, col) == 0) continue;

            /* Nein, Elemente der Rotationsmatrix G werden ermittelt. */
            if(fabs(R(row, col)) >= fabs(R(col, col)))
                t = R(col, col) / R(row, col),
                s = 1 / sqrt(1 + t * t), c = s * t;
            else
                t = R(row, col) / R(col, col),
                c = 1 / sqrt(1 + t * t), s = c * t;

            /* Givens-Rotation Q := G * Q durchfuehren */
            for(k = 0; k < dim; ++k)  
                TMP(col, k) =  Q(row, k) * s + Q(col, k) * c,
                TMP(row, k) =  Q(row, k) * c - Q(col, k) * s;

            for(k = 0; k < dim; ++k)
                Q(col, k) = TMP(col, k),
                Q(row, k) = TMP(row, k);

            /* Givens-Rotation R := G * R durchfuehren */
            for(k = 0; k < dim; ++k)  
                TMP(col, k) =  R(row, k) * s + R(col, k) * c,
                TMP(row, k) =  R(row, k) * c - R(col, k) * s;

            for(k = 0; k < dim; ++k)
                R(col, k) = TMP(col, k),
                R(row, k) = TMP(row, k);
        }
    }

    m_transpose(dim, q);  /* Q = Transp(Q) */

    /* Wenn ein Diagonalelement der Matrix R negativ ist, werden das
     * Vorzeichen der entsprechenden Zeile von R und das Vorzeichen
     * der entsprechenden Spalte von Q invertiert. */
    for(idx = 0; idx < dim; ++idx)
        if(R(idx, idx) < 0)
        {
            m_mul_row(dim, r, idx, -1);
            m_mul_col(dim, q, idx, -1);
        }

    /* Alle Elemente unterhalb der Hauptdiagonalen der Matrix R sind
     * bereits - bis auf Rundungsfehler - gleich null. Nun werden
     * auch noch diese Rundungsfehler beseitigt. */
    for(row = 0; row < dim; ++row)
        for(col = 0; col < row; ++col)
            R(row, col) = 0;

    mat_free(tmp, NULL);
    return MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Es wird die LU-Zerlegung P * MAT = L * U der Matrix "mat" ermittelt.     */
/* Die Ergebnismatrix "lu" enthaelt im oberen Dreieck die Matrix U und      */
/* unterhalb der Hauptdiagonalen L. Die Diagonalelemente von L sind nicht   */
/* in "lu" abgelegt; sie sind alle gleich eins. Die waehrend der Zerlegung  */
/* durchgefuehrten Zeilenvertauschungen sind in der Permutationsmatrix "p"  */
/* wiedergegeben. Mittels "sgn" wird der Wert (-1)^n zurueckgegeben,        */
/* wobei "n" die Anzahl der durchgefuehrten Zeilenvertauschungen ist.       */
/* Die Rueckgabe ist MAT_OK bzw. MAT_ERR_NOINV (wenn "mat" singulaer ist).  */
/* ------------------------------------------------------------------------ */
int m_lu_decomp(size_t dim, const void *mat, void *lu, void *p, double *sgn)
{
    size_t row, col, idx;
    double factor;

    assert(dim > 0);
    assert(mat != NULL);
    assert(lu  != NULL);

    /* "mat", "lu" und "p" muessen unterschiedliche Matrizen sein! */
    assert(mat != lu); assert(mat != p); assert(lu != p);

    /* Matrizen initialisieren. */
    m_copy(dim, lu, mat);
    if(p) m_init(dim, p, MAT_INIT_IDENTITY);
    if(sgn) *sgn = 1;

    /* Zeilen eliminieren. */
    for(idx = 0; idx < dim - 1; ++idx)
    {
        /* Pivotelement suchen. */
        double pivot = LU(idx, idx); size_t pivot_idx = idx;
        for(row = idx + 1; row < dim; ++row)
            if(fabs(LU(row, idx)) > fabs(pivot))
                pivot = LU(row, idx), pivot_idx = row;

        /* Matrix singulaer? */
        if(pivot == 0) return MAT_ERR_NOINV;

        /* Aktuelle Zeile und Pivotzeile tauschen. */
        if(idx != pivot_idx)
        {
            m_swap_row(dim, lu, idx, pivot_idx);
            if(p) m_swap_row(dim,  p, idx, pivot_idx);
            if(sgn) *sgn *= -1;
        }

        /* Zeilen unterhalb "idx" eliminieren. */
        for(row = idx + 1; row < dim; ++row)
        {
            factor = LU(row, idx) = LU(row, idx) / pivot;
            for(col = idx + 1; col < dim; ++col)
                LU(row, col) = LU(row, col) - factor * LU(idx, col);
        }
    }
    return MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Die Koeffizienten des charakteristischen Polynoms der Matrix "mat"       */
/* werden ermittelt und in den Vektor "pol" geschrieben. Im Parameter "dim" */
/* ist die Dimension der Matrix "mat" zu uebergeben. Hinweis: Es werden     */
/* insgesamt (dim + 1) Koeffizienten ermittelt und in "pol" abgelegt.       */
/* Der Ergebnisvektor "pol" muss daher (dim + 1) Elemente besitzen. Die     */
/* Rueckgabe ist MAT_OK oder MAT_ERR_MEMORY (zu wenig freier Speicher).     */
/* ------------------------------------------------------------------------ */
int m_charpol(size_t dim, const void *mat, double *pol)
{
    /* Fuer Details zur Implementierung dieser Funktion siehe:
     * HMMatrix-Dokumentation, "Algorithmus von Samuelson-Berkowitz". */
    size_t i, k, row, col, idx;
    double *pwr = mat_alloc(dim);
    double *z   = mat_alloc(dim);
    double *t   = mat_alloc(dim + 1);

    double *rs1 = vec_alloc(dim);
    double *rs2 = vec_alloc(dim);
    double *q   = vec_alloc(dim + 1);

    assert(dim > 0);
    assert(mat != NULL);
    assert(pol != NULL);

    if(!pwr || !z || !t || !rs1 || !rs2 || !q || !mat || !pol)
    {
        assert(!"m_charpol: out of memory");
        mat_free(pwr, z  , t, NULL);
        vec_free(rs1, rs2, q, NULL);
        return MAT_ERR_MEMORY;
    }

    /* Zeilen 1 + 2, Vektoren "pol" und "q" initialisieren */
    v_fill(dim + 1, q, 0); v_fill(dim + 1, pol, 0);
    pol[0] = 1, pol[1] = -MAT(0, 0);

    /* Zeile 3, Schleife fuer Iterationen */
    for(i = 1; i < dim; ++i)
    {
        /* Zeile 4, Anfangselemente des Vektors "q" belegen */
        q[0] = 1, q[1] = -MAT(i, i);

        /* Zeile 5, restliche Elemente des Vektors "q" belegen */
        for(k = 1; k <= i; ++k)
        {
            /* Zeile 6a, Matrix "Z" belegen */
            for(row = 0; row < dim; ++row)
                for(col = 0; col < dim; ++col)
                    Z(row, col) = (row < i && col < i) * MAT(row, col);

            /* Zeile 6b, Potenz Z^(k-1) per "Square & Multiply" berechnen */
            m_init(dim, pwr, MAT_INIT_IDENTITY);
            for(idx = k - 1; idx != 0; idx /= 2)
            {   /* ...temporaere Variable "T" */
                if(idx % 2) { m_mul(dim, t, pwr, z); m_copy(dim, pwr, t); }
                m_mul(dim, t, z, z); m_copy(dim, z, t); 
            }

            /* Zeile 6c, Produkt Z^(k-1) * s berechnen */
            for(idx = 0; idx < dim; ++idx)
                rs1[idx] = (idx < i) * MAT(idx, i);
            m_vmul(dim, rs2, pwr, rs1);

            /* Zeile 6d, Skalarprodukt -r * [Z^(k-1) * s] berechnen */
            for(idx = 0; idx < dim; ++idx)
                rs1[idx] = (idx < i) * MAT(i, idx);
            q[k + 1] = -v_dot(dim, rs1, rs2);
        }

        /* Zeile 7a, (dim+1)-dimensionale Toeplitzmatrix "T" belegen */
        ++dim;  /* ...Achtung, sonst klappt das Makro T(...) nicht! */
        m_init(dim, t, MAT_INIT_NULL);
        for(col = 0; col < i + 1; ++col)
            for(row = col; row < i + 2; ++row)
                T(row, col) = q[row - col];
        --dim;

        /* Zeile 7b, Koeffizienten des char. Polynoms "pol" berechnen */
        m_vmul(dim + 1, q, t, pol);  /* ...temporaere Variable "q" */
        v_copy(dim + 1, pol, q);
    }

    /* Zeile 8, Algorithmus war erfolgreich */
    mat_free(pwr, z  , t, NULL);
    vec_free(rs1, rs2, q, NULL);
    return MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Mittels zyklischem Jacobi-Verfahren werden alle Eigenwerte und -vektoren */
/* der symmetrischen Matrix "mat" ermittelt. Die Eigenwerte werden auf der  */
/* Hauptdiagonalen der Matrix "eig", die Eigenvektoren in den jeweiligen    */
/* Spalten der Matrix "vec" abgelegt. Die Eigenwerte sind nicht nach ihrer  */
/* Groesse sortiert. Die Rueckgabe der Funktion "m_jacobi" ist entweder     */
/* MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_ITFAIL (wenn das Jacobi-Verfahren    */
/* nach "MAT_MAX_ITER" Iterationen kein Ergebnis mit der Genauigkeit        */
/* "MAT_EPSILON" ermitteln konnte).                                         */
/*                                                                          */  
/* Achtung: Die Funktion "m_jacobi" darf nur fuer symmetrische Matrizen     */
/* "mat" aufgerufen werden. Wenn die Matrix "mat" nicht symmetrisch ist,    */
/* werden keine korrekten Ergebnisse berechnet!                             */
/* ------------------------------------------------------------------------ */
int m_jacobi(size_t dim, const void *mat, void *eig, void *vec)
{
    return m_jacobi2(dim, mat, eig, vec, MAT_EPSILON, MAT_MAX_ITER);
}

/* ------------------------------------------------------------------------ */
/* Mittels zyklischem Jacobi-Verfahren werden alle Eigenwerte und -vektoren */
/* der symmetrischen Matrix "mat" ermittelt. Die Eigenwerte werden auf der  */
/* Hauptdiagonalen der Matrix "eig", die Eigenvektoren in den jeweiligen    */
/* Spalten der Matrix "vec" abgelegt. Die Eigenwerte sind nicht nach ihrer  */
/* Groesse sortiert. Die Rueckgabe der Funktion "m_jacobi2" ist entweder    */
/* MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_ITFAIL (wenn das Jacobi-Verfahren    */
/* innerhalb von "max_iter" Iterationen kein Ergebnis mit der Genauigkeit   */
/* "epsilon" ermitteln konnte).                                             */
/*                                                                          */  
/* Achtung: Die Funktion "m_jacobi2" darf nur fuer symmetrische Matrizen    */
/* "mat" aufgerufen werden. Wenn die Matrix "mat" nicht symmetrisch ist,    */
/* werden keine korrekten Ergebnisse berechnet!                             */
/* ------------------------------------------------------------------------ */
int m_jacobi2(size_t dim, const void *mat, void *eig, void *vec,
    double epsilon, size_t max_iter)
{
    size_t i = 0, stop, p, q, k;
    double *mat1 = (double*)eig,   *mat2 = (double*)vec;
    double *tmp1 = mat_alloc(dim), *tmp2 = mat_alloc(dim);

    assert(dim > 0); 
    assert(eig != vec);
    assert(mat != NULL);
    assert(eig != NULL);
    assert(vec != NULL);

    if(!dim || !mat || !vec || !eig || !tmp1 || !tmp2)
    { mat_free(tmp1, tmp2, NULL); return MAT_ERR_MEMORY; }

    m_copy(dim, mat1, mat); m_copy(dim, tmp1, mat);
    m_init(dim, mat2, MAT_INIT_IDENTITY);
    m_init(dim, tmp2, MAT_INIT_IDENTITY);

    /* Zyklisches Jacobi-Verfahren zur Berechnung der Eigenwerte: Auf der
     * Hauptdiagonalen der Matrix mat1 werden die Eigenwerte abgelegt,
     * in den Spalten der Matrix mat2 die dazugehoerigen Eigenvektoren. */
    do
    {
        for(p = 1, stop = 1; p < dim; ++p)
        {
            for(q = 0; q < p; ++q)
            {
                /* Zu den folgenden Berechnungen siehe: Heinrich Sormann,
                 * Numerische Methoden in der Physik (515.421), Kapitel 7,
                 * Skriptum WS2012/13, Technische Universitaet Graz */
                double nenner, phi, s, c;
                if(MAT1(p, q) * MAT1(p, q) < epsilon * epsilon) continue;
                stop = 0;

                /* Skript, Gleichung 7.34 */
                nenner = MAT1(p, p) - MAT1(q, q), phi = M_PI_4;
                if(fabs(nenner / MAT1(p, q)) > epsilon)
                    phi = .5 * atan(2.0 * MAT1(p, q) / nenner);
                s = sin(phi), c = cos(phi);

                /* Skript, Gleichung 7.28 */
                for(k = 0; k < dim; ++k) if(k != p && k != q)
                    TMP1(p, k) = MAT1(k, p) * c + MAT1(k, q) * s,
                    TMP1(k, p) = TMP1(p, k);

                /* Skript, Gleichung 7.29 */
                for(k = 0; k < dim; ++k) if(k != p && k != q) 
                    TMP1(q, k) = MAT1(k, q) * c - MAT1(k, p) * s,
                    TMP1(k, q) = TMP1(q, k);

                /* Skript, Gleichungen 7.30, 7.31 und 7.32 */
                TMP1(p, p) = MAT1(p, p) * c * c
                    + 2 * MAT1(p, q) * c * s + MAT1(q, q) * s * s;
                TMP1(q, q) = MAT1(q, q) * c * c
                    - 2 * MAT1(p, q) * c * s + MAT1(p, p) * s * s;
                TMP1(p, q) = MAT1(p, q) * (c * c - s * s)
                    + (MAT1(q, q) - MAT1(p, p)) * c * s;
                TMP1(q, p) = TMP1(p, q);

                /* Matrix "mat1" := Matrix "tmp1" */
                for(k = 0; k < dim; ++k)
                    MAT1(k, p) = MAT1(p, k) = TMP1(k, p),
                    MAT1(k, q) = MAT1(q, k) = TMP1(k, q);

                /* Skript, Gleichungen 7.36 und 7.37 */
                for(k = 0; k < dim; ++k)  
                    TMP2(k, p) = MAT2(k, p) * c + MAT2(k, q) * s,
                    TMP2(k, q) = MAT2(k, q) * c - MAT2(k, p) * s;

                /* Matrix "mat2" := Matrix "tmp2" */
                for(k = 0; k < dim; ++k)
                    MAT2(k, p) = TMP2(k, p),
                    MAT2(k, q) = TMP2(k, q);
            }
        }

        /* printf("Jacobi-Zyklus Nr. %u beendet\n", i + 1);
         * m_print(dim, mat1); */

        /* Falls stop == 1 ist, sind alle Nebendiagonalelemente der
         * Matrix mat1 verschwunden und das Verfahren wird beendet.
         * Ansonsten erfolgt ein weiterer Jacobi-Zyklus durch mat1. */
    }
    while((stop == 0) && (++i < max_iter));

    mat_free(tmp1, tmp2, NULL);
    return (i >= max_iter) ? MAT_ERR_ITFAIL : MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Der betragsmaessig groesste Eigenwert der Matrix "mat" und der dazu      */
/* passende Eigenvektor werden ermittelt und ueber die (Zeiger-)Parameter   */
/* "eig" und "vec" an den Aufrufer zurueckgegeben. Der Rueckgabewert der    */
/* Funktion "m_mises2" betraegt MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_ITFAIL  */
/* (falls nach "max_iter" Iterationen kein Ergebnis mit der Genauigkeit     */
/* "epsilon" ermittelt werden konnte).                                      */
/*                                                                          */
/* Achtung: Die Berechnung des Eigenwerts bzw. des Eigenvektors erfolgt     */
/* ueber eine von-Mises-Vektoriteration. Insbesondere bei doppelten, eng    */
/* benachbarten oder komplexen Eigenwerten konvergiert das Verfahren nur    */
/* langsam oder ueberhaupt nicht. In solchen Faellen sollten die Parameter  */
/* "max_iter" und "epsilon" auf geeignete Werte gesetzt werden, um ein      */
/* Einfrieren des Programmablaufs zu vermeiden.                             */
/* ------------------------------------------------------------------------ */
int m_mises2(size_t dim, const void *mat, double *eig, double *vec,
    size_t max_iter, double epsilon)
{
    size_t i = 0;
    double delta, *tmp = vec_alloc(dim);

    /* Die folgende Implementierung basiert auf dem Skript "Computational
     * Physics 1 - Lineare Algebra", Kapitel 10, von Ulrich Hohenester:
     * http://physik.uni-graz.at/~uxh/teaching/compphys1.html */
    assert(dim > 0);
    assert(mat !=  eig); assert(mat !=  vec); assert(eig !=  vec);
    assert(mat != NULL); assert(eig != NULL); assert(vec != NULL);
    if(!tmp) { assert(!"m_mises: out of memory"); return MAT_ERR_MEMORY; }

    /* Zufaelligen Startvektor fuer Iteration generieren; darauf achten
     * dass der Betrag des Vektors nicht zu klein ist, um Rundungsfehler
     * bei der nachfolgenden Normalisierung des Vektors zu vermeiden. */
    do v_init(dim, vec, MAT_INIT_RANDOM); while(v_abs(dim, vec) < .5);
    v_normalize(dim, vec);

    do
    {
        /* Naechster Iterationsschritt */
        m_vmul(dim, tmp, mat, vec);
        *eig = v_dot(dim, tmp, vec);

        if(!*eig)
        {
            /* Entweder gibt es keinen Eigenwert != 0 oder der Startvektor
             * wurde zufaellig exakt senkrecht zum Eigenvektor des groessten
             * Eigenwerts gewaehlt (extrem unwahrscheinlich!). Wir waehlen
             * einfach einen neuen Startvektor und versuchen es nochmals. */
            do v_init(dim, vec, MAT_INIT_RANDOM); while(v_abs(dim, vec) < .5);
            v_normalize(dim, vec);
            delta = 2.0 * epsilon + 1.0;  /* Schleife soll weiterlaufen! */
        }
        else
        {
            assert(v_abs(dim, vec) != 0);

            /* Wenn wir einen Eigenwert (e) bzw. Eigenvektor (v) gefunden
             * haben, dann ist M * v - e * v = 0. Wir berechnen daher den
             * Wert delta = |M * v - e * v| und brechen die Iteration ab,
             * wenn delta <= epsilon erreicht ist. */
            v_fmul(dim, vec, *eig);
            v_fadd(dim, vec, tmp, -1.0);
            delta = v_abs(dim, vec);

            /* Vektor fuer naechsten Iterationsschritt normieren */
            v_copy(dim, vec, tmp);
            v_normalize(dim, vec);
        }
    }
    while((++i < max_iter) && delta > epsilon);
    /* printf("m_mises2: %u Iterationen\n", i); */

    vec_free(tmp, NULL);
    return (i >= max_iter) ? MAT_ERR_ITFAIL : MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Der betragsmaessig groesste Eigenwert der Matrix "mat" und der dazu      */
/* passende Eigenvektor werden ermittelt und ueber die (Zeiger-)Parameter   */
/* "eig" und "vec" an den Aufrufer zurueckgegeben. Der Rueckgabewert der    */
/* Funktion "m_mises" betraegt MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_ITFAIL   */
/* (falls nach "MAT_MAX_ITER" Iterationen kein Ergebnis mit der Genauigkeit */
/* "MAT_EPSILON" ermittelt werden konnte).                                  */
/*                                                                          */
/* Achtung: Die Berechnung des Eigenwerts bzw. des Eigenvektors erfolgt     */
/* ueber eine von-Mises-Vektoriteration. Insbesondere bei doppelten, eng    */
/* benachbarten oder komplexen Eigenwerten konvergiert das Verfahren nur    */
/* langsam oder ueberhaupt nicht. In diesen Faellen ist es sinnvoll, statt  */
/* "m_mises" die Funktion "m_mises2" aufzurufen und dort das Iterationsende */
/* ueber die Parameter "max_iter" bzw. "epsilon" gezielt einzustellen.      */
/* ------------------------------------------------------------------------ */
int m_mises(size_t dim, const void *mat, double *eig, double *vec)
{
    return m_mises2(dim, mat, eig, vec, MAT_MAX_ITER, MAT_EPSILON);
}

/* ------------------------------------------------------------------------ */
/* Einen einzelnen Zeilenvektor der Matrix "mat" zurueckgeben.              */
/* ------------------------------------------------------------------------ */
void m_get_row(size_t dim, size_t row, const void *mat, double *vec)
{
    size_t i;
    assert(dim > 0);
    assert(vec != NULL);
    assert(mat != NULL);

    for(i = 0; i < dim; ++i) vec[i] = MAT(row, i);
}

/* ------------------------------------------------------------------------ */
/* Einen einzelnen Spaltenvektor der Matrix "mat" zurueckgeben.             */
/* ------------------------------------------------------------------------ */
void m_get_col(size_t dim, size_t col, const void *mat, double *vec)
{
    size_t i;
    assert(dim > 0);
    assert(vec != NULL);
    assert(mat != NULL);

    for(i = 0; i < dim; ++i) vec[i] = MAT(i, col);
}

/* ------------------------------------------------------------------------ */
/* Die Hauptdiagonalelemente von "mat" in Form eines Vektors zurueckgeben.  */
/* ------------------------------------------------------------------------ */
void m_get_diag(size_t dim, const void *mat, double *vec)
{
    size_t i;
    assert(dim > 0);
    assert(vec != NULL);
    assert(mat != NULL);

    for(i = 0; i < dim; ++i) vec[i] = MAT(i, i);
}

/* ------------------------------------------------------------------------ */
/* Das maximale Element der Matrix "mat" wird zurueckgegeben.               */
/* ------------------------------------------------------------------------ */
double m_get_max(size_t dim, const void *mat)
{
    double result;
    size_t row, col;
    assert(dim > 0);
    assert(mat != NULL);

    result = MAT(0, 0);
    for(row = 0; row < dim; ++row)
        for(col = 0; col < dim; ++col)
            if(MAT(row, col) > result) result = MAT(row, col);
    return result;
}

/* ------------------------------------------------------------------------ */
/* Das minimale Element der Matrix "mat" wird zurueckgegeben.               */
/* ------------------------------------------------------------------------ */
double m_get_min(size_t dim, const void *mat)
{
    double result;
    size_t row, col;
    assert(dim > 0);
    assert(mat != NULL);

    result = MAT(0, 0);
    for(row = 0; row < dim; ++row)
        for(col = 0; col < dim; ++col)
            if(MAT(row, col) < result) result = MAT(row, col);
    return result;
}

/* ------------------------------------------------------------------------ */
/* Der Vektor "vec" wird initialisiert (je nach Parameter "type").          */
/* ------------------------------------------------------------------------ */
void v_init(size_t dim, double *vec, int type)
{
    size_t i;
    assert(dim > 0);
    assert(vec != NULL);

    switch(type)
    {
    case MAT_INIT_NULL:
        /* Alle Elemente auf null setzen */
        memset(vec, 0, sizeof(double) * dim);
        break;

    case MAT_INIT_RANDOM:
        /* Zufallszahlen zwischen 0...1 erzeugen */
        for(i = 0; i < dim; ++i) vec[i] = (double)rand() / RAND_MAX;
        break;

    case MAT_INIT_ONES:
        /* Alle Elemente auf eins setzen */
        for(i = 0; i < dim; ++i) vec[i] = 1.0;
        break;

    default:
        assert(!"invalid type");
    }
}

/* ------------------------------------------------------------------------ */
/* Alle Elemente des Vektors "vec" werden auf den angegebenen Wert gesetzt. */
/* ------------------------------------------------------------------------ */
void v_fill(size_t dim, double *vec, double value)
{
    size_t i;
    assert(dim > 0);
    assert(vec != NULL);

    for(i = 0; i < dim; ++i) vec[i] = value;
}

/* ------------------------------------------------------------------------ */
/* Alle Elemente des Vektors "src" werden in den Vektor "vec" kopiert.      */
/* ------------------------------------------------------------------------ */
void v_copy(size_t dim, double *vec , const double *src)
{
    assert(dim > 0);
    assert(vec != NULL);
    assert(src != NULL);

    memcpy(vec, src, sizeof(double) * dim);
}

/* ------------------------------------------------------------------------ */
/* Die Elemente des Vektors "vec" werden auf dem Bildschirm ausgegeben.     */
/* Das Format kann ueber die Konstante MAT_PRN_FMT eingestellt werden.      */
/* ------------------------------------------------------------------------ */
void v_print(size_t dim, const double *vec)
{
    size_t idx;
    assert(dim > 0);
    assert(vec != NULL);

    for(idx = 0; idx < dim; ++idx)
    {
        if(idx) printf(" ");
        printf(MAT_PRN_FMT, vec[idx]);
    }            
    printf("\n");
}

/* ------------------------------------------------------------------------ */
/* Die Vektoren "v1" und "v2" werden verglichen. Falls beide Vektoren       */
/* identisch sind, wird MAT_OK zurueckgegeben, andernfalls MAT_ERR_NOTEQ.   */
/* ------------------------------------------------------------------------ */
int v_equal(size_t dim, const double *v1, const double *v2)
{
    assert(dim > 0);
    assert(v1 != NULL);
    assert(v2 != NULL);

    return memcmp(v1, v2, dim * sizeof(double)) ? MAT_ERR_NOTEQ : MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Der Betrag des Vektors "v1" wird berechnet.                              */
/* ------------------------------------------------------------------------ */
double v_abs(size_t dim, const double *v1)
{
    assert(dim > 0);
    assert(v1 != NULL);

    return sqrt(v_dot(dim, v1, v1));
}

/* ------------------------------------------------------------------------ */
/* Der Vektor "v1" wird normalisiert (Rueckgabe == MAT_OK). Hat "v1" die    */
/* Laenge null, bleibt "v1" unveraendert (Rueckgabe == MAT_ERR_VNORM).      */
/* ------------------------------------------------------------------------ */
int v_normalize(size_t dim, double *v1)
{
    double length;
    assert(dim > 0);
    assert(v1 != NULL);

    length = v_abs(dim, v1);
    if(length == 0) return MAT_ERR_VNORM;
    v_fmul(dim, v1, 1.0 / length);
    return MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Das "f-fache" des Vektors "v1" wird zum Vektor "vresult" addiert.        */
/* ------------------------------------------------------------------------ */
void v_fadd(size_t dim, double *vresult, const double *v1, double f)
{
    size_t i;
    assert(dim > 0);
    assert(vresult != NULL);
    assert(v1 != NULL);

    for(i = 0; i < dim; ++i) vresult[i] += f * v1[i];
}

/* ------------------------------------------------------------------------ */
/* Der Vektor "vresult" wird mit "factor" multipliziert.                    */
/* ------------------------------------------------------------------------ */
void v_fmul(size_t dim, double *vresult, double factor)
{
    size_t i;
    assert(dim > 0);
    assert(vresult != NULL);

    for(i = 0; i < dim; ++i) vresult[i] *= factor;
}

/* ------------------------------------------------------------------------ */
/* Das Skalarprodukt der Vektoren "v1" und "v2" wird berechnet.             */
/* ------------------------------------------------------------------------ */
double v_dot(size_t dim, const double *v1, const double *v2)
{
    assert(dim > 0);
    assert(v1 != NULL);
    assert(v2 != NULL);

    /* Juli 2017: Um eine bessere Optimierung/Parallelisierung durch
     * den Compiler zu ermoeglichen, wurde die Schleife "zerlegt"...  */
    size_t i;
    double sum[4] = { 0 };
    for(i = 0; i + 3 < dim; i += 4, v1 += 4, v2 += 4)
        sum[0] += v1[0] * v2[0],
        sum[1] += v1[1] * v2[1],
        sum[2] += v1[2] * v2[2],
        sum[3] += v1[3] * v2[3];
    for(; i != dim; ++i, ++v1, ++v2) sum[0] += v1[0] * v2[0];
    return sum[0] + sum[1] + sum[2] + sum[3];
}

/* ------------------------------------------------------------------------ */
/* Das Kreuzprodukt der dreidimensionalen (!) Vektoren "v1" und "v2" wird   */
/* berechnet und im Vektor "vresult" abgespeichert.                         */
/* ------------------------------------------------------------------------ */
void v_cross(double *vresult, const double *v1, const double *v2)
{
    assert(vresult != NULL);
    assert(v1 != NULL);
    assert(v2 != NULL);

    /* Faktoren und Produkt muessen unterschiedliche Vektoren sein! */
    assert(vresult != v1); assert(vresult != v2);

    vresult[0] = v1[1] * v2[2] - v1[2] * v2[1];
    vresult[1] = v1[2] * v2[0] - v1[0] * v2[2];
    vresult[2] = v1[0] * v2[1] - v1[1] * v2[0];
}

/* ------------------------------------------------------------------------ */
/* Das maximale Element des Vektors "vec" wird zurueckgegeben.              */
/* ------------------------------------------------------------------------ */
double v_get_max(size_t dim, const double *vec)
{
    double result;
    size_t idx;
    assert(dim > 0);
    assert(vec != NULL);

    result = vec[0];
    for(idx = 1; idx < dim; ++idx)
        if(vec[idx] > result) result = vec[idx];
    return result;
}

/* ------------------------------------------------------------------------ */
/* Das minimale Element des Vektors "vec" wird zurueckgegeben.              */
/* ------------------------------------------------------------------------ */
double v_get_min(size_t dim, const double *vec)
{
    double result;
    size_t idx;
    assert(dim > 0);
    assert(vec != NULL);

    result = vec[0];
    for(idx = 1; idx < dim; ++idx)
        if(vec[idx] < result) result = vec[idx];
    return result;
}

/* ------------------------------------------------------------------------ */
/* Das Produkt aus der Matrix "mat" und dem Vektor "v" wird berechnet und   */
/* im Vektor "vresult" abgelegt.                                            */
/* ------------------------------------------------------------------------ */
void m_vmul(size_t dim, double *vresult, const void *mat, const double *v)
{
    size_t row;
    assert(dim > 0);
    assert(vresult != NULL);
    assert(mat != NULL);
    assert(v != NULL);

    /* "v" und "vresult" muessen unterschiedliche Vektoren sein! */
    assert(vresult != v);
    for(row = 0; row < dim; ++row)
        vresult[row] = v_dot(dim, &MAT(row, 0), v);
}

/* ------------------------------------------------------------------------ */
/* Der Wert des Polynoms a[0] x^n + a[1] x^(n-1) + a[2] x^(n-2)... + a[n]   */
/* wird fuer den angegebenen komplexen Wert "x" berechnet. Die Anzahl der   */
/* komplexen Koeffizienten ist im Parameter "dim" zu uebergeben.            */
/* ------------------------------------------------------------------------ */
HM_COMPLEX v_cpolyval(size_t dim, const HM_COMPLEX *a, HM_COMPLEX x)
{
    size_t i;
    HM_COMPLEX result = {{ 0, 0 }};  /* Berechnung mittels Horner-Schema */
    for(i = 0; i < dim; ++i) result = cadd(cmul(result, x), a[i]);
    return result;
}

/* ------------------------------------------------------------------------ */
/* Der Wert des Polynoms a[0] x^n + a[1] x^(n-1) + a[2] x^(n-2)... + a[n]   */
/* wird fuer den angegebenen reellen Wert "x" berechnet. Die Anzahl der     */
/* reellen Koeffizienten ist im Parameter "dim" zu uebergeben.              */
/* ------------------------------------------------------------------------ */
double v_dpolyval(size_t dim, const double *a, double x)
{
    size_t i;
    double result = 0;  /* Berechnung mittels Horner-Schema */
    for(i = 0; i < dim; ++i) result *= x, result += a[i];
    return result;
}

/* ------------------------------------------------------------------------ */
/* Alle reellen und komplexen Nullstellen eines Polynoms werden ermittelt   */
/* und im Vektor "roots" abgelegt. Die komplexen Koeffizienten des Polynoms */
/* sind in a[0], ..., a[n] zu uebergeben, deren Anzahl ist im Parameter     */
/* "dim" zu uebergeben. Hinweis: Ein Polynom mit z Koeffizienten besitzt    */
/* genau (z-1) Nullstellen. Der Ergebnisvektor "roots" hat also ein Element */
/* weniger als Vektor "a". Rueckgabe: MAT_OK, MAT_ERR_ITFAIL (Algorithmus   */
/* konvergiert nicht) oder MAT_ERR_DIM (falls dim < 2 oder a[0] == 0).      */
/* ------------------------------------------------------------------------ */
int v_croots(size_t dim, const HM_COMPLEX *a, HM_COMPLEX *roots)
{
    HM_COMPLEX wdk_init = {{ MAT_WDK_INIT }};
    return v_croots2(dim, a, roots, MAT_EPSILON, MAT_MAX_ITER, wdk_init);
}

/* ------------------------------------------------------------------------ */
/* Siehe v_croots(). Bei dieser Variante kann zusaetzlich die gewuenschte   */
/* Genauigkeit, die maximale Anzahl der Iterationen sowie der Startwert     */
/* fuer das verwendete Weierstrass-Durand-Kerner-Iterationsverfahren        */
/* angegeben werden. Rueckgabe: MAT_OK, MAT_ERR_ITFAIL (der Algorithmus     */
/* konvergiert nicht) oder MAT_ERR_DIM (falls dim < 2 oder a[0] == 0).      */
/* ------------------------------------------------------------------------ */
int v_croots2(size_t dim, const HM_COMPLEX *a, HM_COMPLEX *roots,
    double epsilon, size_t max_iter, HM_COMPLEX wdk_init)
{
    /* Zur Implementierung des Iterationsverfahrens: DurandKerner method,
     * http://en.wikipedia.org/w/index.php?title=Durand%E2%80%93Kerner_
     * method&oldid=551301621 (last visited July 13, 2013). */
    HM_COMPLEX init = wdk_init;
    size_t i, j, k;

    assert(a && roots);
    assert(dim > 1 && c_abs(a[0]) != 0);
    assert(epsilon > 0 && max_iter > 0 && c_abs(wdk_init) > 0);
    if(dim < 2 || c_abs(a[0]) == 0) return MAT_ERR_DIM;

    for(i = 0; i + 1 < dim; ++i)
        init = cmul(roots[i] = init, wdk_init);

    for(i = 0; i < max_iter; ++i)
    {
        double eps = 0;
        for(j = 0; j + 1 < dim; ++j)
        {
            HM_COMPLEX tmp = cdiv(v_cpolyval(dim, a, roots[j]), a[0]);
            for(k = 0; k + 1 < dim; ++k)
                if(j != k) tmp = cdiv(tmp, csub(roots[j], roots[k]));
            tmp = csub(roots[j], tmp);

            /* roots[j] = alte und tmp = neue Naeherung fuer j-te Nullst. */
            eps = MAT_MAX(c_abs(csub(roots[j], tmp)), eps);
            roots[j] = tmp;
        }
        if(eps < epsilon) return MAT_OK;
    }

    assert(!"croots: iteration failed");
    return MAT_ERR_ITFAIL;
}

/* ------------------------------------------------------------------------ */
/* Alle reellen und komplexen Nullstellen eines Polynoms werden ermittelt   */
/* und im Vektor "roots" abgelegt. Die reellen Koeffizienten des Polynoms   */
/* sind in a[0], ..., a[n] zu uebergeben, deren Anzahl ist im Parameter     */
/* "dim" zu uebergeben. Hinweis: Ein Polynom mit z Koeffizienten besitzt    */
/* genau (z-1) Nullstellen. Der Ergebnisvektor "roots" hat also ein Element */
/* weniger als Vektor "a". Rueckgabe: MAT_OK, MAT_ERR_ITFAIL (Algorithmus   */
/* konvergiert nicht), MAT_ERR_DIM (falls dim < 2 oder a[0] == 0) oder      */
/* MAT_ERR_MEMORY (freier Speicher reicht nicht aus).                       */
/* ------------------------------------------------------------------------ */
int v_droots(size_t dim, const double *a, HM_COMPLEX *roots)
{
    HM_COMPLEX wdk_init = {{ MAT_WDK_INIT }};
    return v_droots2(dim, a, roots, MAT_EPSILON, MAT_MAX_ITER, wdk_init);
}

/* ------------------------------------------------------------------------ */
/* Siehe v_droots(). Bei dieser Variante kann zusaetzlich die gewuenschte   */
/* Genauigkeit, die maximale Anzahl der Iterationen sowie der Startwert     */
/* fuer das verwendete Weierstrass-Durand-Kerner-Iterationsverfahren        */
/* angegeben werden. Rueckgabe: MAT_OK, MAT_ERR_ITFAIL (der Algorithmus     */
/* konvergiert nicht), MAT_ERR_DIM (falls dim < 2 oder a[0] == 0) oder      */
/* MAT_ERR_MEMORY (freier Speicher reicht nicht aus).                       */
/* ------------------------------------------------------------------------ */
int v_droots2(size_t dim, const double *a, HM_COMPLEX *roots,
    double epsilon, size_t max_iter, HM_COMPLEX wdk_init)
{
    int ok;
    size_t i;
    HM_COMPLEX *cpoly = (HM_COMPLEX*)malloc(sizeof(HM_COMPLEX) * dim);
    if(!cpoly) { assert(!"droots: out of memory"); return MAT_ERR_MEMORY; }

    for(i = 0; i < dim; ++i) RE(cpoly[i]) = a[i], IM(cpoly[i]) = 0;
    ok = v_croots2(dim, cpoly, roots, epsilon, max_iter, wdk_init);
    free(cpoly);
    return ok;
}

/* ------------------------------------------------------------------------ */
/* Die FFT (Fast Fourier Transform, schnelle Fouriertransformation) wird    */
/* fuer den angegebenen Vektor "f" mit "n" komplexen Elementen berechnet.   */
/* Achtung: Die Elementanzahl "n" muss eine Zweierpotenz sein, sonst wird   */
/* die Funktion mit dem Fehlercode MAT_ERR_DIM abgebrochen. Das Ergebnis    */
/* der FFT wird wiederum im Vektor "f" zurueckgegeben. Die Elemente des     */
/* Vektors "f" werden also durch den Aufruf von "v_cfft" ueberschrieben.    */
/* Moegliche Rueckgabewerte: MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_DIM.       */
/* ------------------------------------------------------------------------ */
int v_cfft(size_t n, HM_COMPLEX *f)
{
    /* Details zur Implementierung: H. R. Schwarz, Zuerich,
     * Elementare Darstellung der schnellen Fouriertransformation,
     * Computing 18, 107-116 (1977) */
    int ret = MAT_OK;
    HM_COMPLEX *w = 0;

    BEGIN_BLOCK
    {
        size_t n_ok, i, j, p, ngr, ig, inddif;
        assert(f && "v_cfft: ivalid parameter, null pointer");
        if(!f) { ret = MAT_ERR_MEMORY; break; }

        /* Die Anzahl der Vektorelemente muss eine Zweierpotenz sein */
        n_ok = (n >= 2) && !(n & (n - 1));
        assert(n_ok && "v_cfft: invalid vector size");
        if(!n_ok) { ret = MAT_ERR_DIM;  break; }
        p = (size_t)(log(1. * n) / log(2.) + .5);

        /* Speicherplatz fuer FFT-Berechnung reservieren */
        w = (HM_COMPLEX*)malloc(n / 2 * sizeof(HM_COMPLEX));
        assert(w && "v_cfft: out of memory");
        if(!w) { ret = MAT_ERR_MEMORY; break; }
    
        for(i = 0; i < n / 2; ++i)
            RE(w[i]) = cos(-2. * M_PI * i / n),
            IM(w[i]) = sin(-2. * M_PI * i / n);
    
        inddif = n / 2;
        for(ig = ngr = 1; ig <= p; ++ig, ngr *= 2, inddif /= 2)
        {
            HM_COMPLEX yh;
            size_t inda, igr, ip, i1, i2;
            for(igr = 1, inda = 0; igr <= ngr; ++igr, inda += 2 * inddif)
                for(i = ip = 0; i < inddif; ++i, ip += ngr)
                    i1 = inda + i, i2 = i1 + inddif,
                    yh = csub(f[i1], f[i2]),
                    f[i1] = cadd(f[i1], f[i2]),
                    f[i2] = cmul(yh, w[ip]);
        }

        for(j = 0; j < n; ++j)
        {
            size_t k = 0, jh, kp;
            for(i = 1, jh = j, kp = n / 2; i <= p; ++i, jh /= 2, kp /= 2)
                k = k + (jh % 2) * kp;
            if(k > j) c_swap(f + j, f + k);
        }
    }
    END_BLOCK;

    free(w);
    return ret;
}

/* ------------------------------------------------------------------------ */
/* Die FFT (Fast Fourier Transform, schnelle Fouriertransformation) wird    */
/* fuer den angegebenen Vektor "f" mit "n" reellen Elementen berechnet.     */
/* Achtung: Die Elementanzahl "n" muss eine Zweierpotenz sein, sonst wird   */
/* die Funktion mit dem Fehlercode MAT_ERR_DIM abgebrochen. Das Ergebnis    */
/* der FFT wird in den Vektor "y" geschrieben, welcher Platz fuer "n"       */
/* komplexe Elemente des Typs HM_COMPLEX bieten muss. Rueckgabewerte von    */
/* "v_dfft" koennen sein: MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_DIM.          */
/* ------------------------------------------------------------------------ */
int v_dfft(size_t n, const double *f, HM_COMPLEX *y)
{
    size_t i;
    assert(n > 0 && f != NULL && y != NULL);

    for(i = 0; i < n; ++i) RE(y[i]) = f[i], IM(y[i]) = 0;
    return v_cfft(n, y);
}

/* ------------------------------------------------------------------------ */
/* Die IFFT (Inverse Fast Fourier Transform) wird fuer den Vektor "y" mit   */
/* "n" komplexen Elementen berechnet. Achtung: Die Elementanzahl "n" muss   */
/* eine Zweierpotenz sein, sonst wird "v_ifft" mit dem Fehler MAT_ERR_DIM   */
/* abgebrochen. Das Ergebnis der IFFT wird im Vektor "y" zurueckgegeben,    */
/* die Elemente des Vektors "y" werden also durch den Aufruf von "v_ifft"   */
/* ueberschrieben. Rueckgabewerte: MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_DIM. */
/* ------------------------------------------------------------------------ */
int v_ifft(size_t n, HM_COMPLEX *y)
{
    size_t i;
    assert(n > 0 && y != NULL);

    for(i = 0; i < n; ++i) RE(y[i]) /= (double)n, IM(y[i]) /= -(double)n;
    return v_cfft(n, y);
}

/* ------------------------------------------------------------------------ */
/* Das lineare Gleichungssystem (QR) * x = y wird geloest. Vor dem Aufruf   */
/* von "lin_solve_qr" muss zunaechst eine QR-Zerlegung der Matrix erfolgen, */
/* zum Beispiel mittels "m_qr_decomp". (Alternativ kann auch die Funktion   */
/* "lin_solve" aufgerufen werden, wo beide Schritte - QR-Zerlegung und die  */
/* die Loesung des Gleichungssystems - gemeinsam durchgefuehrt werden.)     */
/* Der Rueckgabewert betraegt MAT_OK, falls das Gleichungssystem geloest    */
/* werden konnte, ansonsten MAT_ERR_NOINV.                                  */
/* ------------------------------------------------------------------------ */
int lin_solve_qr(size_t dim, double *x,
    const void *q, const void *r, const double *y)
{
    size_t row, col, i, j;

    assert(dim  >  0);
    assert(x !=    y); assert(q !=    r);
    assert(x != NULL); assert(y != NULL);
    assert(q != NULL); assert(r != NULL);

    /* x := Transp(q) * y */
    for(row = 0; row < dim; ++row)
        for(col = 0, x[row] = 0; col < dim; ++col)
            x[row] += Q(col, row) * y[col];

    /* Gleichungssystem durch Rueckwaerts-Einsetzen loesen */
    for(i = dim - 1; i != (size_t)-1; --i)
    {
        if(!R(i, i)) return MAT_ERR_NOINV;
        for(j = i - 1; j != (size_t)-1; --j)
            if(R(j, i) != 0)
                x[j] -= R(j, i) / R(i, i) * x[i], R(j, i) = 0;
        x[i] /= R(i, i), R(i, i) = 1;
    }
    return MAT_OK;
}

/* ------------------------------------------------------------------------ */
/* Das lineare Gleichungssystem A * x = y wird geloest. Es wird zunaechst   */
/* eine QR-Zerlegung der Matrix durchgefuehrt und anschliessend die Loesung */
/* des Gleichungssystems durch Rueckwaerts-Einsetzen bestimmt. Moegliche    */
/* Rueckgabewerte sind MAT_OK, MAT_ERR_MEMORY und MAT_ERR_NOINV. Bitte die  */
/* Kommentare zur Funktion "lin_solve_qr" ebenfalls beachten.               */
/* ------------------------------------------------------------------------ */
int lin_solve(size_t dim, double *x, const void *a, const double *y)
{
    int ok;
    double *q = mat_alloc(dim), *r = mat_alloc(dim);

    assert(dim > 0);
    assert(a != NULL);
    assert(x != y);

    /*  Es wird zunaechst die QR-Zerlegung von A ermittelt. */
    if(q == NULL || r == NULL || m_qr_decomp(dim, a, q, r) != MAT_OK)
    { mat_free(q, r, NULL); return MAT_ERR_MEMORY; }

    /* Nun kann das Gleichungssystem geloest werten. */
    ok = lin_solve_qr(dim, x, q, r, y);
    mat_free(q, r, NULL);
    return ok;
}

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird im CSV-Format in die bereits geoeffnete Datei      */
/* "file" geschrieben. Die Datei bleibt nach dem Schreiben geoeffnet.       */
/* Rueckgabewert == MAT_OK, falls die Matrix erfolgreich ausgegeben wurde   */
/* oder MAT_ERR_FWRITE, falls ein Fehler bei der Ausgabe aufgetreten ist.   */
/* ------------------------------------------------------------------------ */
int m_writecsv2(size_t dim, const void *mat, FILE *file)
{
    size_t row, col;
    int ok = 0;

#if MAT_USE_LOCALE
    char *oldloc = str_duplicate_(setlocale(LC_NUMERIC, 0)); 
    if(oldloc) setlocale(LC_NUMERIC, "");
    assert(oldloc != NULL);
#endif

    assert(dim > 0);
    assert(mat != NULL);
    assert(file != NULL);

    for(row = 0; row < dim; ++row)
    {
        fprintf(file, MAT_CSV_FMT, MAT(row, 0));  /* erste Spalte */
        for(col = 1; col < dim; ++col)  /* restliche Spalten mit ; */
            fprintf(file, ";" MAT_CSV_FMT, MAT(row, col));
        ok = fprintf(file, "\n");
    }

#if MAT_USE_LOCALE
    setlocale(LC_NUMERIC, oldloc);
    free(oldloc);
#endif

    return (ok > 0) ? MAT_OK : MAT_ERR_FWRITE;
}

/* ------------------------------------------------------------------------ */
/* Es wird die Datei "filename" erstellt und die Matrix "mat" im CSV-Format */
/* gespeichert. Rueckgabewert == MAT_OK, falls die Ausgabe erfolgreich war, */
/* MAT_ERR_FOPEN bzw. MAT_ERR_FWRITE, falls ein Fehler aufgetreten ist.     */
/* ------------------------------------------------------------------------ */
int m_writecsv(size_t dim, const void *mat, const char *filename)
{
    int ok;
    FILE* f;
    assert(dim > 0);
    assert(mat != NULL);
    assert(filename != NULL);

    f = fopen(filename, "w");
    if(!f) return MAT_ERR_FOPEN;
    ok = m_writecsv2(dim, mat, f);
    fclose(f);
    return ok;
}

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird aus der bereits geoeffneten CSV-Datei "file"       */
/* eingelesen, danach bleibt die Datei geoeffnet. Falls die Matrix korrekt  */
/* eingelesen wurde, ist der Rueckgabewert == MAT_OK, sonst MAT_ERR_FORMAT. */
/* ------------------------------------------------------------------------ */
int m_readcsv2(size_t dim, void *mat, FILE *file)
{
    enum { NUM, SEP, RDY, ERR } state = NUM;
    size_t row = 0, col = 0;
    int ok;
    char separator;
    double value;

#if MAT_USE_LOCALE
    char *oldloc = str_duplicate_(setlocale(LC_NUMERIC, 0));
    if(oldloc) setlocale(LC_NUMERIC, "");
    assert(oldloc != NULL);
#endif

    assert(dim > 0);
    assert(mat != NULL);
    assert(file != NULL);

    do
    {
        if(state == NUM)
        {
            ok = fscanf(file, "%lf", &value), state = ERR;
            if(!ok) break;
            MAT(row, col++) = value, state = SEP;
            if(col == dim) col = 0, ++row, state = NUM;
            if(row == dim) state = RDY;
        }
        if(state == SEP)
        {
            ok = fscanf(file, " %c", &separator);
            state = (ok && separator == ';') ? NUM : ERR;
        }
    }
    while(state != RDY && state != ERR);

#if MAT_USE_LOCALE
    setlocale(LC_NUMERIC, oldloc);
    free(oldloc);
#endif

    return (state == RDY) ? MAT_OK : MAT_ERR_FORMAT;
}

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird aus der CSV-Datei "filename" eingelesen. Falls     */
/* die Matrix korrekt eingelesen wurde, ist der Rueckgabewert == MAT_OK,    */
/* ansonsten MAT_ERR_FORMAT oder MAT_ERR_FOPEN.                             */
/* ------------------------------------------------------------------------ */
int m_readcsv(size_t dim, void *mat, const char *filename)
{
    int ok;
    FILE* f;
    assert(dim > 0);
    assert(mat != NULL);
    assert(filename != NULL);

    f = fopen(filename, "r");
    if(!f) return MAT_ERR_FOPEN;
    ok = m_readcsv2(dim, mat, f);
    fclose(f);
    return ok;
}
